from collections import OrderedDict
import gym
from gym import (spaces,logger)
import numpy as np

from protein_folding_environment.base_environment import ProteinFoldingBaseEnv
from utils.movement_utils import move_to_new_state_3d
from utils.plotting_utils import plot_3D_foleded_protein

class ProteinFolding3DEnv(ProteinFoldingBaseEnv):
    def __init__(self, seq):
        """
        Initializes a new instance of the `ProteinFolding3DEnv` class.

        Parameters:
        -----------
        seq : str
            The protein sequence to be folded.

        Returns:
        --------
        None
        """
        super().__init__(seq)
        self.reset()
        self.action_space = spaces.Discrete(start=1, n=6)
        self.observation_space = spaces.Box(low=0, high=5,
                                            shape=(len(self.seq) - 2,),
                                            dtype=int)

    def step(self, action):
        """
        Performs one step of the protein folding simulation.

        Parameters:
        -----------
        action : int
            The action to be taken by the agent.

        Returns:
        --------
        tuple
            The new observation, the reward, whether the episode is done, and additional information.
        """
        if not self.action_space.contains(action):
            raise ValueError("%r (%s) invalid" % (action, type(action)))

        self.last_action = action

        previous = list(self.state.keys())[-1]
        previous2 = list(self.state.keys())[-1]
        next_state = move_to_new_state_3d(
            previous,
            previous2,
            action
        )
        idx = len(self.state)
        if next_state is None or next_state in self.state:
            return (None, None, False, False, {})

        self.actions.append(action)
        try:
            self.state.update({next_state: self.seq[idx]})
        except IndexError:
            logger.error('All molecules have been placed! Nothing can be added to the protein chain.')
            raise

        return self.get_observation_info(next_state)

    def reset(self):
        """
        Resets the environment to the initial state.

        Returns:
        --------
        numpy.ndarray
            The initial observation of the environment.
        """
        self.actions = []
        self.state = OrderedDict(
            {
                (0, 0, 0): self.seq[0],
                (1, 0, 0): self.seq[1],
            }
        )
        self.done = len(self.seq) == 2
        obs = self.observe()
        return obs

    def render(self, mode='human', display_mode="draw",
               pause_t=0.0, save_fig=False, save_path="",
               score=2022, optima_idx=0):
        """
        Renders the current state of the environment.

        Parameters:
        -----------
        mode : str, optional
            The rendering mode. 'human' for displaying in the console, 'rgb_array' for returning an RGB array.
        display_mode : str, optional
            The display mode. 'draw' for drawing the environment using ASCII art, 'plot' for plotting the environment
            using Matplotlib.
        pause_t : float, optional
            The amount of time to pause the rendering.
        save_fig : bool, optional
            Whether to save the rendering as a figure.
        save_path : str, optional
            The path to save the rendering as a figure.
        score : int, optional
            The current score of the environment.
        optima_idx : int, optional
            The index of the optimal score.

        Returns:
        --------
        None
        """
        if mode == "human":
            plot_3D_foleded_protein(
                list(self.state.items()),
            )

    def _get_adjacent_coords(self, coords):
        """
        Returns a dictionary containing adjacent coordinates for a given set of coordinates.

        Parameters:
        -----------
        coords: tuple
            The coordinates (x, y, z) for which adjacent coordinates are to be found.

        Returns:
        --------
        dict
            A dictionary containing adjacent coordinates as values and their respective directions as keys.

        """
        x, y, z = coords
        adjacent_coords = {
            0: (x - 1, y, z),
            1: (x, y - 1, z),
            2: (x, y + 1, z),
            3: (x + 1, y, z),
            4: (x, y, z - 1),
            5: (x, y, z + 1)
        }
        return adjacent_coords

    def _compute_free_energy(self, chain):
        """
        Computes the free energy of a given chain.

        Parameters:
        -----------
        chain: OrderedDict
            An ordered dictionary containing the coordinates and corresponding amino acids in a chain.

        Returns:
        --------
        int
            The total free energy of the chain.

        """
        path = list(chain.items())
        total_energy = 0
        for index in range(0, len(path)):
            for jndex in range(index, len(path)):
                if abs(index - jndex) >= 2:
                    current_amino_acid_i = path[index][1]
                    current_amino_acid_j = path[jndex][1]
                    current_place_i = path[index][0]
                    current_place_j = path[jndex][0]
                    x_i = current_place_i[0]
                    y_i = current_place_i[1]
                    z_i = current_place_i[2]
                    x_j = current_place_j[0]
                    y_j = current_place_j[1]
                    z_j = current_place_j[2]
                    if current_amino_acid_i == 'H' and current_amino_acid_j == 'H' and (
                            abs(x_i - x_j) + abs(y_i - y_j) + abs(z_i - z_j) == 1):
                        total_energy += 1
        return total_energy





from collections import OrderedDict
from gym import (spaces)

from protein_folding_environment.base_environment import ProteinFoldingBaseEnv
from utils.plotting_utils import plot_2D_folded_protein


class ProteinFolding2DEnv(ProteinFoldingBaseEnv):
    def __init__(self, seq):
        """
        Initializes an instance of the `ProteinFolding2DEnv` class with a protein sequence `seq`.

        Args:
        - seq (str): the protein sequence to be folded.

        Calls the `__init__` method of the base class `ProteinFoldingBaseEnv` and sets the following instance variables:
        - self.actions (list): a list of actions taken so far.
        - self.state (OrderedDict): an ordered dictionary representing the current state of the protein chain.
        """
        super().__init__(seq)
        self.reset()

    def reset(self):
        """
        Resets the environment to its initial state.

        Returns:
        - obs (numpy array): the current observation of the environment.

        Sets the following instance variables:
        - self.actions (list): an empty list of actions taken so far.
        - self.state (OrderedDict): an ordered dictionary representing the current state of the protein chain.
        - self.done (bool): a flag indicating whether the episode is over.
        - obs (numpy array): the current observation of the environment.
        """
        self.actions = []
        self.state = OrderedDict(
            {
                (0, 0): self.seq[0],
                (1, 0): self.seq[1],
            }
        )
        self.done = len(self.seq) == 2
        obs = self.observe()
        return obs

    def render(self, mode='human'):
        """
        Renders the current state of the protein chain.

        Args:
        - mode (str): the rendering mode to use.

        If `mode` is "human", calls `plot_2D_folded_protein` to render the current state of the protein chain.
        """
        if mode == "human":
            plot_2D_folded_protein(
                list(self.state.items()),
            )



from collections import OrderedDict

import gym
from gym import (spaces, logger)
import numpy as np

from protein_folding_environment.base_environment_2d import ProteinFolding2DEnv
from utils.movement_utils import move_to_new_state_lrf



class ProteinFoldingLRF2DEnv(ProteinFolding2DEnv):
    def __init__(self, seq):
        """
        Initializes a new instance of the ProteinFoldingLRF2DEnv class.

        Parameters:
        seq (str): The amino acid sequence of the protein to fold.

        """
        super().__init__(seq)
        self.action_space = spaces.Discrete(start=1, n=3)
        self.observation_space = spaces.Box(low=0, high=3,
                                            shape=(len(self.seq) - 2,),
                                            dtype=int)

    def step(self, action):
        """
        Executes one step within the environment.

        Parameters:
        action (int): The action to execute.

        Returns:
        obs (numpy.ndarray): The current observation of the environment.
        reward (float): The reward earned by the agent.
        done (bool): Whether the environment is in a terminal state.
        info (dict): Optional additional information.

        """
        if not self.action_space.contains(action):
            raise ValueError("%r (%s) invalid" % (action, type(action)))

        previous = list(self.state.keys())[-1]
        new_move_direction, next_state = move_to_new_state_lrf(self.move_direction,previous,action)
        self.move_direction = new_move_direction
        idx = len(self.state)
        if next_state is None or next_state in self.state:
            return (None, None, False, False, {})
        self.actions.append(action)
        try:
            self.state.update({next_state: self.seq[idx]})
        except IndexError:
            logger.error('All molecules have been placed! Nothing can be added to the protein chain.')
            raise

        return self.get_observation_info(next_state)

    def reset(self):
        """
        Resets the environment to an initial state.

        Returns:
        obs (numpy.ndarray): The current observation of the environment.

        """
        obs = super().reset()
        self.move_direction = (0, 1)
        return obs

    def _get_adjacent_coords(self, coords):
        """
        Gets the adjacent coordinates for a given set of coordinates.

        Parameters:
        coords (tuple): The coordinates for which to get the adjacent coordinates.

        Returns:
        adjacent_coords (dict): A dictionary containing the adjacent coordinates.

        """
        x, y = coords
        adjacent_coords = {
            0: (x - 1, y),
            1: (x, y - 1),
            2: (x, y + 1),
            3: (x + 1, y),
        }
        return adjacent_coords


import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.lines as mlines
import numpy as np
from matplotlib.ticker import (MultipleLocator, AutoMinorLocator, MaxNLocator)
from sklearn.metrics import euclidean_distances


def plot_print_rewards_stats(rewards_all_episodes,
                             show_every,
                             args,
                             mode="show",
                             save_path="./rewards"):
    """
    Plot and print the rewards statistics over the course of training.

    Args:
        rewards_all_episodes (list): List of rewards for each episode.
        show_every (int): Number of episodes after which to print the statistics.
        args (Namespace): Namespace object containing the parsed arguments.
        mode (str): Plotting mode: "show" to display the plot, "save" to save it to a file (default: "show").
        save_path (str): Path to the directory to save the plot (default: "./rewards").

    Returns:
        None
    """
    # unpack the args
    seq = args.seq
    seed = args.seed
    num_episodes = args.num_episodes

    # Calculate and print the average reward per show_every episodes
    rewards_per_N_episodes = np.split(
        np.array(rewards_all_episodes),
        num_episodes
    )
    count = show_every

    # for plotting
    aggr_ep_rewards = {'ep': [], 'avg': [], 'max': [], 'min': []}

    print("\n********Stats per {} episodes********\n".format(show_every))
    for r in rewards_per_N_episodes:
        # print(count, "avg: ", str(sum(r/show_every)))
        # print(count, "min: ", str(min(r)))
        # print(count, "max: ", str(max(r)))

        aggr_ep_rewards['ep'].append(count)
        aggr_ep_rewards['avg'].append(sum(r / show_every))
        aggr_ep_rewards['min'].append(min(r))
        aggr_ep_rewards['max'].append(max(r))

        count += show_every

    # Width, height in inches.
    # default: [6.4, 4.8]
    fig_width = 6.4
    fig_height = 4.8
    # adjust the height of the histogram
    if np.array(rewards_all_episodes).max() - np.array(rewards_all_episodes).min() > 10:
        fig_width = 6.5
        fig_height = 6.5
    fig, subplot = plt.subplots(figsize=(fig_width, fig_height))
    # Be sure to only pick integer tick locations
    subplot.yaxis.set_major_locator(MaxNLocator(integer=True))
    # subplot.yaxis.set_major_locator(MultipleLocator(1))
    subplot.yaxis.set_minor_locator(MultipleLocator(1))

    subplot.set_xlabel('Episode Index')
    subplot.set_ylabel('Episode Reward')

    subplot.plot(aggr_ep_rewards['ep'], aggr_ep_rewards['avg'], label="average rewards")
    subplot.plot(aggr_ep_rewards['ep'], aggr_ep_rewards['max'], label="max rewards")
    subplot.plot(aggr_ep_rewards['ep'], aggr_ep_rewards['min'], label="min rewards")

    # Put a legend below current axis
    subplot.legend(loc='upper center', bbox_to_anchor=(0.5, -0.15),
                   fancybox=True, shadow=True, ncol=3)

    # split the seq into chunks of 10 for the matplotlib title
    chunks, chunk_size = len(seq), 10
    seq_title_list = [
        seq[i:i + chunk_size] + "\n" for i in range(0, chunks, chunk_size)
    ]

    plt.grid(True, which="major", lw=1.2, linestyle='-')
    plt.grid(True, which="minor", lw=0.8, linestyle='--')
    plt.tight_layout()
    if mode == "show":
        plt.show()
    elif mode == "save":
        # save the pdf fig with seq name
        plt.savefig("{}-{}-{}-{}.png".format(
            save_path,  # "./xxx"
            seq,
            num_episodes,
            seed,
        ))
    plt.close()


def plot_2D_folded_protein(labelled_conf):
    """
    input:
        labelled_conf:
            transformed file sequence of xy coords with state:
            ((x,y), 'H|P')
    output:
        plot.show
    """

    plt.rc('axes', labelsize=18)
    plt.rc('xtick', labelsize=18)
    plt.rc('ytick', labelsize=18)

    fig = plt.figure()
    subplot = fig.add_subplot()
    subplot.title.set_text("Folded Protein")
    x = [t[0][0] for t in labelled_conf]
    y = [t[0][1] for t in labelled_conf]

    H_points = [t[0] for t in labelled_conf if t[1] == 'H']
    P_points = [t[0] for t in labelled_conf if t[1] == 'P']

    max_xval = np.max(x)
    max_yval = np.max(y)

    total_max = max(max_xval, max_yval)
    min_xval = np.min(x)
    min_yval = np.min(y)

    total_min = min(min_xval, min_yval)
    subplot.set_xlim(total_min - 0.1, total_max + 0.1)
    subplot.set_ylim(total_min - 0.1, total_max + 0.1)

    subplot.set_aspect('equal')  # , adjustable='box')

    subplot.xaxis.set_major_locator(ticker.MultipleLocator(1))
    subplot.yaxis.set_major_locator(ticker.MultipleLocator(1))

    subplot.plot(
        x, y,
        color='cornflowerblue',
        linewidth=4,
        label="backbone",
    )

    subplot.plot(
        [h[0] for h in H_points],
        [h[1] for h in H_points],
        'o',
        color='royalblue',
        markersize=14,
        label="H",
    )

    for index in range(0, len(labelled_conf)):
        for jndex in range(index, len(labelled_conf)):
            if abs(index - jndex) >= 2:
                current_amino_acid_i = labelled_conf[index][1]
                current_amino_acid_j = labelled_conf[jndex][1]
                current_place_i = labelled_conf[index][0]
                current_place_j = labelled_conf[jndex][0]
                x_i = current_place_i[0]
                y_i = current_place_i[1]
                x_j = current_place_j[0]
                y_j = current_place_j[1]
                if current_amino_acid_i == 'H' and current_amino_acid_j == 'H' and (
                        abs(x_i - x_j) + abs(y_i - y_j) == 1):
                    subplot.plot([x_i, x_j], [y_i, y_j], '--', color='mediumblue')
                    subplot.plot([x_i, x_j], [y_i, y_j], 'o', color='mediumblue', markersize=14, )

    subplot.plot(
        [p[0] for p in P_points],
        [p[1] for p in P_points],
        'o',
        color='orange',
        fillstyle='none',
        markersize=14,
        label="P",
    )
    plt.show()
def plot_3D_foleded_protein(labelled_conf):
    """
    input:
        labelled_conf:
            transformed file sequence of xy coords with state:
            ((x,y,z), 'H|P')
    output:
        plot.show
    """

    plt.rc('axes', labelsize=25)
    plt.rc('xtick', labelsize=21)
    plt.rc('ytick', labelsize=21)
    x = [t[0][0] for t in labelled_conf]
    y = [t[0][1] for t in labelled_conf]
    z = [t[0][2] for t in labelled_conf]

    H_points = [t[0] for t in labelled_conf if t[1] == 'H']
    P_points = [t[0] for t in labelled_conf if t[1] == 'P']

    fig = plt.figure()
    subplot = fig.add_subplot(projection='3d')
    subplot.title.set_text("Folded Protein")

    max_xval = np.max(x)
    max_yval = np.max(y)
    max_zval = np.max(z)
    total_max = max(max_xval, max_yval, max_zval)
    min_xval = np.min(x)
    min_yval = np.min(y)
    min_zval = np.min(z)
    total_min = min(min_xval, min_yval, min_zval)
    subplot.set_xlim(total_min, total_max)
    subplot.set_ylim(total_min, total_max)
    subplot.set_zlim(total_min, total_max)

    subplot.grid(linewidth=0.6, linestyle=':')

    subplot.set_aspect('equal')  # , adjustable='box')

    subplot.xaxis.set_major_locator(ticker.MultipleLocator(1))
    subplot.yaxis.set_major_locator(ticker.MultipleLocator(1))
    subplot.zaxis.set_major_locator(ticker.MultipleLocator(1))

    subplot.set_xlabel("Ox")
    subplot.set_ylabel("Oy")
    subplot.set_zlabel("Oz")

    subplot.plot(
        x, y, z,
        color='cornflowerblue',
        linewidth=4,
        label="backbone",
    )

    subplot.plot(
        [h[0] for h in H_points],
        [h[1] for h in H_points],
        [h[2] for h in H_points],
        'o',
        color='royalblue',
        markersize=14,
        label="H",
    )

    subplot.plot(
        [p[0] for p in P_points],
        [p[1] for p in P_points],
        [p[2] for p in P_points],
        'o',
        color='orange',
        fillstyle='none',
        markersize=14,
        label="P",
    )

    for index in range(0, len(labelled_conf)):
        for jndex in range(index, len(labelled_conf)):
            if abs(index - jndex) >= 2:
                current_amino_acid_i = labelled_conf[index][1]
                current_amino_acid_j = labelled_conf[jndex][1]
                current_place_i = labelled_conf[index][0]
                current_place_j = labelled_conf[jndex][0]
                x_i = current_place_i[0]
                y_i = current_place_i[1]
                z_i = current_place_i[2]
                x_j = current_place_j[0]
                y_j = current_place_j[1]
                z_j = current_place_j[2]
                if current_amino_acid_i == 'H' and current_amino_acid_j == 'H' and (
                        abs(x_i - x_j) + abs(y_i - y_j) + abs(z_i - z_j) == 1):
                    subplot.plot([x_i, x_j], [y_i, y_j], [z_i, z_j], '--', color='mediumblue')
                    subplot.plot([x_i, x_j], [y_i, y_j], [z_i, z_j], 'o', color='mediumblue', markersize=14, )

    plt.show()

def moving_average(a, n=2):
    """
    Computes the moving average of a list of numbers.

    Parameters:
    a (list): A list of numbers.
    n (int): The size of the window to compute the moving average.

    Returns:
    numpy.ndarray: A numpy array with the moving average of the input list.
    """
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n
def plot_moving_avg(scores, n=2, mode="show", save_path="./avg-rewards"):
    """
    Plots the moving average of a list of scores.

    Parameters:
    scores (list): A list of scores.
    n (int): The size of the window to compute the moving average. Default is 2.
    mode (str): The mode of the plot: "show" to show the plot or "save" to save it to disk. Default is "show".
    save_path (str): The path to save the plot. Default is "./avg-rewards".

    Returns:
    None.
    """
    print("means = ", scores.mean())
    plt.xlabel('Episode')
    plt.ylabel('Reward')
    plt.plot(moving_average(scores, n=n))

    if mode == "show":
        plt.show()
    elif mode == "save":
        # save the png fig with window size
        plt.savefig("{}-{}.png".format(
            save_path,  # "./xxx"
            n
        ))
    plt.close()


def plot_loss(episodes, losses):
    """
    Plots the loss values of a reinforcement learning model.

    Parameters:
    episodes (list): A list of episode numbers.
    losses (list): A list of loss values.

    Returns:
    None.
    """
    fig, subplot = plt.subplots()

    subplot.set_xlabel('Episode Index')
    subplot.set_ylabel('Episode Loss')

    subplot.plot(episodes, losses, label="Losses")

    plt.show()
    plt.close()



from collections import OrderedDict

import gym
from gym import (spaces, logger)
import numpy as np

from protein_folding_environment.base_environment_2d import ProteinFolding2DEnv
from utils.movement_utils import move_to_new_state_2d


class ProteinFoldingSquareEnv(ProteinFolding2DEnv):
    def __init__(self, seq):
        """
        Initializes an instance of the `ProteinFoldingSquareEnv` class with a protein sequence `seq`.

        Args:
        - seq (str): the protein sequence to be folded.

        Calls the `__init__` method of the base class `ProteinFolding2DEnv` and sets the following instance variables:
        - self.action_space (Discrete): a discrete action space representing the four possible moves in a square lattice.
        - self.observation_space (Box): a box observation space representing the current chain of amino acids in the protein sequence.
        """
        super().__init__(seq)
        self.action_space = spaces.Discrete(start=1, n=4)
        self.observation_space = spaces.Box(low=0, high=3,
                                            shape=(len(self.seq) - 2,),
                                            dtype=int)

    def step(self, action):
        """
        Advances the environment by one step given an action.

        Args:
        - action (int): the action to take.

        Returns:
        - obs (numpy array): the current observation of the environment.
        - reward (float): the reward value for the current state of the environment.
        - done (bool): a flag indicating whether the episode is over.
        - False (bool): a flag indicating whether the environment has successfully reset.
        - info (dict): a dictionary containing information about the current state of the environment.

        If `action` is not a valid action, raises a `ValueError`.
        Calculates the next state of the protein chain using `move_to_new_state_2d()` based on the current state and the proposed action and updates the environment's state with the next state.
        If `next_state` is None or already in the state dictionary, returns `(None, None, False, False, {})`.
        Appends `action` to the list of actions taken so far and calls `self.get_observation_info(next_state)` to get the current observation, reward, done flag, and info dictionary based on the current state of the environment and the proposed next state.
        """
        if not self.action_space.contains(action):
            raise ValueError("%r (%s) invalid" % (action, type(action)))
        previous = list(self.state.keys())[-1]
        previous2 = list(self.state.keys())[-2]
        next_state = move_to_new_state_2d(
            previous,
            previous2,
            action
        )
        idx = len(self.state)
        if next_state is None or next_state in self.state:
            return (None, None, False, False, {})
        self.actions.append(action)
        try:
            self.state.update({next_state: self.seq[idx]})
        except IndexError:
            logger.error('All molecules have been placed! Nothing can be added to the protein chain.')
            raise
        return self.get_observation_info(next_state)

    def _get_adjacent_coords(self, coords):
        """
        Returns a dictionary containing the adjacent coordinates of a given set of coordinates in a square lattice.

        Args:
        - coords (tuple): the coordinates to get adjacent coordinates for.

        Returns:
        - adjacent_coords (dict): a dictionary containing the adjacent coordinates of `coords`.

        Calculates the adjacent coordinates of `coords` in a square lattice and returns them as a dictionary.
        """
        x, y = coords
        adjacent_coords = {
            0: (x - 1, y),
            1: (x, y - 1),
            2: (x, y + 1),
            3: (x + 1, y),
        }
        return adjacent_coords



from gym.envs.registration import register
from .base_environment import ProteinFoldingBaseEnv
from .base_environment_2d import ProteinFolding2DEnv
from .environement_triunghiular import ProteinFoldingTriangularEnv
from .environment_2d import ProteinFoldingSquareEnv
from .environment_3d import ProteinFolding3DEnv
from .environment_LRF import ProteinFoldingLRF2DEnv

#register all custom environments
register(
    id='ProteinFoldingSquareEnv',
    entry_point='protein_folding_environment.environment_2d:ProteinFoldingSquareEnv',
    max_episode_steps=100000,
)

register(
    id='ProteinFolding3DEnv',
    entry_point='protein_folding_environment.environment_3d:ProteinFolding3DEnv',
    max_episode_steps=100000,
)

register(
    id='ProteinFoldingLRF2DEnv',
    entry_point='protein_folding_environment.environment_LRF:ProteinFoldingLRF2DEnv',
    max_episode_steps=100000,
)

register(
    id='ProteinFoldingTriangularEnv',
    entry_point='protein_folding_environment.environement_triunghiular:ProteinFoldingTriangularEnv',
    max_episode_steps=100000,
)



import numpy as np

ACTION_TO_MEANING_2D = {1: (-1, 0), 2: (1, 0), 3: (0, -1), 4: (0, 1)}
ACTION_TO_MEANING_3D = {1: (-1, 0, 0), 2: (1, 0, 0), 3: (0, -1, 0), 4: (0, 1, 0), 5: (0, 0, -1), 6: (0, 0, 1)}
ACTION_TO_DIRECTION = {
    (-1, 0): {'L': (0, -1), 'R': (0, 1), 'F': (-1, 0)},
    (1, 0): {'L': (0, 1), 'R': (0, -1), 'F': (1, 0)},
    (0, -1): {'L': (1, 0), 'R': (-1, 0), 'F': (0, -1)},
    (0, 1): {'L': (0, -1), 'R': (1, 0), 'F': (0, 1)},
}
ACTION_TO_LRF = {
    1: 'L',
    2: 'R',
    3: 'F'
}
LRF_TO_ACTION = {
    'L': 1,
    'R': 2,
    'F': 3
}
ACTION_TO_MEANING_TRIANGULAR = {
    1: (-1, 0), 2: (1, 0), 3: (0, -1), 4: (0, 1), 5: (1, 1), 6: (-1, -1)
}


def move_to_new_state_lrf(previous_move_direction, state, action):
    """
    Computes the new state of the agent given the current state, action, and previous move direction in the LRF 2D environment.

    Args:
        previous_move_direction (tuple): The previous move direction of the agent.
        state (tuple): The current state of the agent.
        action (int): The chosen action to take.

    Returns:
        tuple: A tuple containing the new move direction and the new state.

    """
    if action not in {1, 2, 3}:
        return

    action_lrf = ACTION_TO_LRF[action]
    current_move_direction = ACTION_TO_DIRECTION[previous_move_direction][action_lrf]
    x1, y1 = state

    new_state = (x1 + current_move_direction[0], y1 + current_move_direction[1])

    return current_move_direction, new_state


def move_to_new_state_3d(p1, p2, move_direction):
    """
    Computes the new state of the agent given the current state, action, and previous move direction in the 3D environment.

    Args:
        p1 (tuple): The current position of the agent.
        p2 (tuple): The previous position of the agent.
        move_direction (int): The chosen action to take.

    Returns:
        tuple: A tuple containing the new state.

    """
    if move_direction not in {1, 2, 3, 4, 5, 6}:
        return
    x1, y1, z1 = p1

    new_state = (x1 + (ACTION_TO_MEANING_3D[move_direction])[0], y1 + (ACTION_TO_MEANING_3D[move_direction])[1],
                 z1 + (ACTION_TO_MEANING_3D[move_direction])[2])
    if p2 == new_state:
        return
    return new_state


def move_to_new_state_2d(p1, p2, move_direction):
    """
    Computes the new state of the agent given the current state, action, and previous move direction in the 2D environment.

    Args:
        p1 (tuple): The current position of the agent.
        p2 (tuple): The previous position of the agent.
        move_direction (int): The chosen action to take.

    Returns:
        tuple: A tuple containing the new state.

    """
    if move_direction not in {1, 2, 3, 4}:
        return

    x1, y1 = p1

    new_state = (x1 + (ACTION_TO_MEANING_2D[move_direction])[0], y1 + (ACTION_TO_MEANING_2D[move_direction])[1])
    if p2 == new_state:
        return
    return new_state


def move_to_new_state_triangular(p1, p2, move_direction):
    """
    Computes the new state of the agent given the current state, action, and previous move direction in the triangular environment.

    Args:
        p1 (tuple): The current position of the agent.
        p2 (tuple): The previous position of the agent.
        move_direction (int): The chosen action to take.

    Returns:
        tuple: A tuple containing the new state.

    """
    if move_direction not in {1, 2, 3, 4, 5, 6}:
        return

    x1, y1 = p1
    new_state = (
        x1 + (ACTION_TO_MEANING_TRIANGULAR[move_direction])[0], y1 + (ACTION_TO_MEANING_TRIANGULAR[move_direction])[1])
    if p2 == new_state:
        return
    return new_state



from collections import OrderedDict

import gym
from gym import (spaces, utils, logger)
import numpy as np


class ProteinFoldingBaseEnv(gym.Env):
    def __init__(self, seq):
        """
        Initializes an instance of the `ProteinFoldingBaseEnv` class with a protein sequence `seq`.

        Args:
        - seq (str): the protein sequence to be folded.

        Initializes instance variables:
        - self.seq (str): the protein sequence to be folded.
        - self.is_trapped (bool): a flag indicating whether the protein chain is trapped in a local minimum.
        - self.done (bool): a flag indicating whether the episode is over.
        """
        self.seq = seq.upper()
        self.is_trapped = False
        self.done = False
        if len(self.seq) <= 2:
            return

    def observe(self):
        """
        Returns a numpy array representation of the current chain of amino acids in the protein sequence.

        Returns:
        - native_obs (numpy array): an array representing the current chain of amino acids in the protein sequence.
        """
        action_chain = self.actions
        native_obs = np.zeros(shape=(len(self.seq) - 2,), dtype=int)
        for i, item in enumerate(action_chain):
            native_obs[i] = item
        return native_obs

    def _compute_reward(self):
        """
        Computes the reward for the current state of the environment.

        Returns:
        - reward (float): the reward value for the current state of the environment.

        Computes `curr_reward` by calling `_compute_free_energy(self.state)` and returns the following based on the state of the protein chain:
        - -0.01 if `self.is_trapped` is True.
        - `curr_reward` if `self.done` is True.
        - 0 otherwise.
        """
        curr_reward = self._compute_free_energy(self.state)
        if self.is_trapped:
            return -0.01
        elif self.done:
            return curr_reward
        else:
            return 0

    def _compute_free_energy(self, chain):
        """
        Computes the free energy of the protein chain represented by `chain`.

        Args:
        - chain (OrderedDict): the protein chain to calculate free energy for.

        Returns:
        - total_energy (float): the free energy of the protein chain.

        Computes `total_energy` as the sum of pairwise interactions between all pairs of hydrophobic residues that are separated by at least two other residues.
        """
        path = list(chain.items())
        total_energy = 0
        for index in range(0, len(path)):
            for jndex in range(index, len(path)):
                if abs(index - jndex) >= 2:
                    current_amino_acid_i = path[index][1]
                    current_amino_acid_j = path[jndex][1]
                    current_place_i = path[index][0]
                    current_place_j = path[jndex][0]
                    x_i = current_place_i[0]
                    y_i = current_place_i[1]
                    x_j = current_place_j[0]
                    y_j = current_place_j[1]
                    if current_amino_acid_i == 'H' and current_amino_acid_j == 'H' and (
                            abs(x_i - x_j) + abs(y_i - y_j) == 1):
                        total_energy += 1
        return total_energy


    def get_observation_info(self, next_state):
        """
        Returns the current observation, reward, done flag, and info dictionary based on the current state of the environment and the proposed next state.

        Args:
        - next_state (OrderedDict): the proposed next state of the protein chain.

        Returns:
        - obs (numpy array): the current observation of the environment.
        - reward (float): the reward value for the current state of the environment.
        - done (bool): a flag indicating whether the episode is over.
        - False (bool): a flag indicating whether the environment has successfully reset.
        - info (dict): a dictionary containing information about the current state of the environment.

        Calls `self.observe()` to get the current observation of the environment and sets the following based on the current state of the protein chain and the proposed next state:
        - is_trapped (bool): a flag indicating whether the protein chain is trapped in a local minimum.
        - self.is_trapped (bool): updates the value of `self.is_trapped`.
        - self.done (bool): updates the value of `self.done`.
        - reward (float): computes the reward for the current state by calling `_compute_reward()` and updates the value of `self.done`.
        - info (dict): a dictionary containing the current chain length, sequence length, actions taken so far, whether the chain is trapped, and the current state of the protein chain.
        """
        is_trapped = False
        if len(self.state) < len(self.seq):
            if set(self._get_adjacent_coords(next_state).values()).issubset(self.state.keys()):
                is_trapped = True
        obs = self.observe()
        self.is_trapped = is_trapped
        self.done = len(self.state) == len(self.seq) or is_trapped
        reward = self._compute_reward()
        info = {
            'chain_length': len(self.state),
            'seq_length': len(self.seq),
            'actions': [str(i) for i in self.actions],
            'is_trapped': is_trapped,
            'state_chain': self.state,
        }
        return obs, reward, self.done, False, info


    def seed(self, seed=None):
        """
        Seeds the environment's random number generator and the action space's random number generator with a given seed value or with a randomly generated seed value if none is provided.

        Args:
        - seed (int, optional): the seed value to use for the random number generator.

        Returns:
        - [seed] (list): a list containing the generated seed value.
        """
        self.np_random, seed = utils.seeding.np_random(seed)
        self.action_space.seed(seed)
        np.random.seed(seed)

        return [seed]


import re

from Bio import SeqIO

# we ignore the X marking as they are unknown amino acids
ATOM_TO_HYDROPHOBICITY = {
    'G': 'H',
    'A': 'H',
    'V': 'H',
    'L': 'H',
    'I': 'H',
    'P': 'H',
    'F': 'H',
    'M': 'H',
    'W': 'H',
    'S': 'P',
    'T': 'P',
    'C': 'P',
    'N': 'P',
    'Q': 'P',
    'Y': 'P',
    'X': 'P',
    'R': 'P',
    'H': 'P',
    'K': 'P',
    'D': 'P'
}

d3to1 = {'CYS': 'C', 'ASP': 'D', 'SER': 'S', 'GLN': 'Q', 'LYS': 'K',
         'ILE': 'I', 'PRO': 'P', 'THR': 'T', 'PHE': 'F', 'ASN': 'N',
         'GLY': 'G', 'HIS': 'H', 'LEU': 'L', 'ARG': 'R', 'TRP': 'W',
         'ALA': 'A', 'VAL': 'V', 'GLU': 'E', 'TYR': 'Y', 'MET': 'M',
         }


class InputHandler:
    @staticmethod
    def extract_hydropolar_sequence_from_pdb(file_name):
        """
                Extracts the hydropolar sequence of an amino acid chain from a PDB file.

                Args:
                    file_name (str): The path to the PDB file.

                Returns:
                    str: The hydropolar sequence of the amino acid chain.
                """
        amino_acid_sequence = ""
        previous_chain = ""
        with open(file_name) as f:
            for matched_line in re.findall('^SEQRES.*$', f.read(), re.MULTILINE):
                tokens = matched_line.split()
                chain = tokens[2]
                if previous_chain != "" and chain != previous_chain:
                    print(
                        "Warning: The model works only with single-chain proteins. "
                        "The input will be truncated to use the first chain. ")
                    break
                tokens = tokens[4:]
                amino_acid_sequence = ''.join([d3to1[t] if t in d3to1.keys() else '' for t in tokens])
                amino_acid_sequence = ''.join([ATOM_TO_HYDROPHOBICITY[t] for t in amino_acid_sequence])
                previous_chain = chain

        return amino_acid_sequence

    @staticmethod
    def extract_hydropolar_sequence_from_fasta(file_name):
        """
                Extracts the hydropolar sequence of an amino acid chain from a FASTA file.

                Args:
                    file_name (str): The path to the FASTA file.

                Returns:
                    str: The hydropolar sequence of the amino acid chain.
                """
        amino_acid_sequence = ""
        index = 0
        for seq_record in SeqIO.parse(file_name, "fasta"):
            if index > 0:
                print(
                    "Warning: The model works only with single-chain proteins. "
                    "The input will be truncated to use the first chain. ")
                break
            sequence = seq_record.seq
            amino_acid_sequence = ''.join([ATOM_TO_HYDROPHOBICITY[t] for t in sequence])
            index += 1

        return amino_acid_sequence

seq = InputHandler.extract_hydropolar_sequence_from_fasta('/Users/ancaioanamuscalagiu/Documents/licenta/ProteinFolding/known_proteins/fasta/rcsb_pdb_1A1P.fasta')

print(seq)


from setuptools import setup, find_packages

VERSION = '0.0.1'
DESCRIPTION = 'Package with Reinforcement Learning Environments for Protein Folding'
LONG_DESCRIPTION = 'Python Library implementing various Gym Environments for solving the Protein Folding Problem using Reinforcement Learning'

setup(
    name="protein-folding-gym-utils",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="Muscalagiu Anca Ioana",
    author_email="ancamuscalagiu1@gmail.com",
    license='MIT',
    packages=find_packages(),
    install_requires=['bio==1.5.9',
'biopython==1.81',
'biothings-client==0.3.0',
'certifi==2022.12.7',
'charset-normalizer==3.1.0',
'cloudpickle==2.2.1',
'contourpy==1.0.7',
'cycler==0.11.0',
'fonttools==4.39.3',
'gprofiler-official==1.0.0',
'gym==0.26.2',
'gym-notices==0.0.8',
'importlib-metadata==6.6.0',
'importlib-resources==5.12.0',
'joblib==1.2.0',
'kiwisolver==1.4.4',
'matplotlib==3.7.1',
'mygene==3.2.2',
'numpy==1.23.5',
'packaging==23.1',
'pandas==2.0.1',
'Pillow==9.5.0',
'platformdirs==3.3.0',
'pooch==1.7.0',
'pyparsing==3.0.9',
'python-dateutil==2.8.2',
'pytz==2023.3',
'requests==2.28.2',
'scikit-learn==1.2.2',
'scipy==1.10.1',
'six==1.16.0',
'threadpoolctl==3.1.0',
'tqdm==4.65.0',
'tzdata==2023.3',
'urllib3==1.26.15',
'zipp==3.15.0',
],
    keywords=['protein', 'folding', 'reinforcement'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ]
)
